from dataclasses import dataclass
from typing import *

from algoritmia.schemes.bt_scheme import DecisionSequence, bt_solve, bt_vc_solve
from algoritmia.schemes.bt_scheme import ScoredDecisionSequence, bt_min_solve

Decision = int
Solution = tuple[Decision, ...]
State = tuple[int, int]


def coin_change_solve_naif(v: tuple[int, ...], Q: int) -> Iterable[Solution]:
    def calc(ds: tuple[int, ...]) -> int:
        return sum(ds[i] * v[i] for i in range(len(ds)))

    class CoinChangeDS(DecisionSequence):
        def is_solution(self) -> bool:
            q = calc(self.decisions())
            return len(self) == len(v) and q == Q

        def successors(self) -> Iterable["CoinChangeDS"]:
            n = len(self)
            if n < len(v):
                q = calc(self.decisions())
                pending = Q - q
                for num_coins in range(pending // v[n] + 1):
                    yield self.add_decision(num_coins)

    initial_ds = CoinChangeDS()
    return bt_solve(initial_ds)


def coin_change_solve(v: tuple[int, ...], Q: int) -> Iterable[Solution]:
    @dataclass
    class Extra:
        pending: int

    class CoinChangeDS(DecisionSequence):
        def is_solution(self) -> bool:
            return len(self) == len(v) and self.extra.pending == 0

        def successors(self) -> Iterable["CoinChangeDS"]:
            n = len(self)
            if n < len(v):
                for num_coins in range(self.extra.pending // v[n] + 1):
                    pending2 = self.extra.pending - num_coins * v[n]
                    yield self.add_decision(num_coins, Extra(pending2))

    initial_ds = CoinChangeDS(Extra(Q))
    return bt_solve(initial_ds)


def coin_change_vc_solve(v: tuple[int, ...], Q: int) -> Iterable[Solution]:
    @dataclass
    class Extra:
        pending: int

    class CoinChangeDS(DecisionSequence):
        def is_solution(self) -> bool:
            return len(self) == len(v) and self.extra.pending == 0

        def successors(self) -> Iterable["CoinChangeDS"]:
            n = len(self)
            if n < len(v):
                for num_coins in range(self.extra.pending // v[n] + 1):
                    pending2 = self.extra.pending - num_coins * v[n]
                    yield self.add_decision(num_coins, Extra(pending2))

        def state(self):
            return len(self), self.extra.pending

    initial_ds = CoinChangeDS(Extra(Q))
    return bt_vc_solve(initial_ds)


def coin_change_opt_solve(v: tuple[int, ...], Q: int) -> Iterable[Solution]:
    @dataclass
    class Extra:
        pending: int

    class CoinChangeDS(ScoredDecisionSequence):
        def is_solution(self) -> bool:
            return len(self) == len(v) and self.extra.pending == 0

        def successors(self) -> Iterable["CoinChangeDS"]:
            n = len(self)
            if n < len(v):
                for num_coins in range(self.extra.pending // v[n] + 1):
                    pending2 = self.extra.pending - num_coins * v[n]
                    yield self.add_decision(num_coins, Extra(pending2))

        def state(self):
            return len(self), self.extra.pending

        def score(self):
            return sum(self.decisions())

    initial_ds = CoinChangeDS(Extra(Q))
    return bt_min_solve(initial_ds)


# Programa principal ---------------------------------
if __name__ == "__main__":
    coins, quantity = (1, 2, 5), 7
    for solve in [coin_change_solve, coin_change_vc_solve, coin_change_opt_solve]:
        print(solve.__name__)
        for sol in solve(coins, quantity):
            print(sol)
        print()



from algoritmia.datastructures.graphs import Digraph, WeightingFunction
from algoritmia.utils import infinity


def solve(g: Digraph, d: WeightingFunction, s, t) -> tuple[int, list[int]]:
    def L(v) -> int:
        if v == s:
            return 0
        if v not in mem:
            mem[v] = min((L(u) + d(u, v), u) for u in g.preds(v))
        return mem[v][0]

    mem = {}
    score = L(t)
    v = t
    sol = [v]
    while v != s:
        _, v = mem[v]
        sol.append(v)
    sol.reverse()
    return score, sol


def solve2(g: Digraph, d: WeightingFunction, s, t) -> tuple[int, list[int]]:
    def L(v, n) -> int:
        if v == s:
            return 0
        if n == 0:
            return infinity
        if (v, n) not in mem:
            mem[v, n] = min((L(u, n - 1) + d(u, v), (u, n - 1)) for u in g.preds(v))
        return mem[v, n][0]

    mem = {}
    score = L(t, len(g.V) - 1)
    if score == infinity:
        return score, []
    v = t
    n = len(g.V) - 1
    sol = [v]
    while v != s:
        _, (v, n) = mem[v, n]
        sol.append(v)
    sol.reverse()
    return score, sol


if __name__ == '__main__':
    d = {(0, 1): 1, (0, 3): 50, (1, 2): 10, (2, 0): 2, (2, 3): -3, (3, 1): -2, (3, 2): 100, (3, 4): 4}
    my_edges = d.keys()
    g = Digraph(E=my_edges)
    wf = WeightingFunction((e for e in d.items()))

    print(solve2(g, wf, 3, 2))



from algoritmia.utils import infinity


def solve(q: int, v: list[int]) -> tuple[int, list[int]]:
    def L(n, q):
        if n == 0:
            return 0 if q == 0 else infinity
        if (n, q) not in mem:
            mem[n, q] = min(
                [(L(n - 1, q - d * v[n - 1]) + d, (n - 1, q - d * v[n - 1], d)) for d in range(q // v[n - 1] + 1)])
        return mem[n, q][0]

    mem = {}
    score = L(len(v), q)
    n = len(v)
    sol = []
    while n > 0:
        _, (n, q, d) = mem[n, q]
        sol.append(d)
    sol.reverse()
    return score, sol


if __name__ == '__main__':
    v, q = [1, 2, 5, 10], 11
    print("Instance:", q, v)
    print("Solution:", solve(q, v))



from typing import TypeVar, Iterable

from algoritmia.datastructures.graphs import IGraph, UndirectedGraph, WeightingFunction

from algoritmia.algorithms.traversers import bf_edge_traverser, dijkstra_edge_traverser, dijkstra_metric_edge_traverser

Vertex = TypeVar('Vertex')
Edge = tuple[Vertex, Vertex]
Path = list[Vertex]


def shortest_path_unweighted_graph(g: IGraph[Vertex], v_source: Vertex, v_target: Vertex) -> Path:
    edges = bf_edge_traverser(g, v_source)
    return path_recover(edges, v_target)


def shortest_path_positive_weighted_graph(g: IGraph[Vertex], d: WeightingFunction,
                                          v_source: Vertex, v_target: Vertex) -> Path:
    edges = dijkstra_edge_traverser(g, d, v_source)
    return path_recover(edges, v_target)


def shortest_path_metric_graph(g: IGraph[Vertex], d: WeightingFunction,
                               v_source: Vertex, v_target: Vertex) -> Path:
    edges = dijkstra_metric_edge_traverser(g, d, v_source, v_target)
    return path_recover(edges, v_target)


# Slide 37
def path_recover(edges: Iterable[Edge], v: Vertex) -> Path:
    # Creates backpointer dictionary (bp)
    bp = {}
    for o, d in edges:
        bp[d] = o
        if d == v:  # I have all I need
            break
    # Recover the path jumping back
    path = [v]
    while v != bp[v]:
        v = bp[v]
        path.append(v)
    # reverse the path
    path.reverse()
    return path


if __name__ == '__main__':
    # -------------
    # Output: [(0, 0), (0, 1), (1, 1), (1, 2), (0, 2), (0, 3), (1, 3)]

    my_edges = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)), ((1, 1), (1, 2)),
                ((2, 0), (2, 1)), ((2, 1), (2, 2)), ((2, 2), (2, 3)), ((0, 1), (1, 1)),
                ((0, 2), (1, 2)), ((0, 3), (1, 3)), ((1, 1), (2, 1)), ((1, 2), (2, 2))]
    my_graph = UndirectedGraph(E=my_edges)
    v_initial = (0, 0)
    v_final = (1, 3)

    path = shortest_path_unweighted_graph(my_graph, v_initial, v_final)
    print('shortest_path_unweighted_graph:', path)

    # ---------------
    # Output: ['Madrid', 'Venturada', 'Aranda de Duero', 'Lerma', 'Burgos', 'Miranda del Ebro', 'Armiñón', 'Basauri', 'Bilbao']

    from algoritmia.data.iberia import iberia, km

    path = shortest_path_positive_weighted_graph(iberia, km, 'Madrid', 'Bilbao')
    print('shortest_path_positive_weighted_graph:', path)

    from algoritmia.data.iberia import iberia2d, km2d, coords2d, coords2d_inv

    pathc = shortest_path_metric_graph(iberia2d, km2d, coords2d['Madrid'], coords2d['Bilbao'])
    path = [coords2d_inv[c] for c in pathc]
    print('shortest_path_metric_graph:', path)



from typing import *


def coin_change_solve_naif(v: tuple[int, ...], Q: int) -> Optional[list[int]]:
    res = []
    for v_i in v:
        res.append(Q // v_i)
        Q = Q % v_i
    if Q == 0:
        return res
    return None


def coin_change_solve(v: tuple[int, ...], Q: int) -> Optional[list[int]]:
    # creamos vector de índices para recorrer 'valores' de mayor a menor valor
    indices_ordenados = sorted(range(len(v)), key=lambda i: -v[i])

    res = [0] * len(v)
    for i in indices_ordenados:
        res[i] = Q // v[i]
        Q = Q % v[i]
    if Q == 0:
        return res
    return None


if __name__ == '__main__':
    print(coin_change_solve_naif((1, 2, 5, 10), 6))  # [6, 0, 0, 0]
    print(coin_change_solve_naif((2, 5, 10), 7))  # None
    print(coin_change_solve_naif((1, 9, 15), 19))  # [19, 0, 0]
    print(coin_change_solve_naif((2, 9, 15), 10))  # [5, 0, 0]
    print()
    print(coin_change_solve((1, 2, 5, 10), 6))  # [1, 0, 1, 0]
    print(coin_change_solve((2, 5, 10), 7))  # [1, 1, 0]]
    print(coin_change_solve((1, 9, 15), 19))  # [4, 0, 1]
    print(coin_change_solve((2, 9, 15), 10))  # None



from algoritmia.examples.greedy.graph_coloring_greedy import coloring_solve
from algoritmia.viewers.graph2d_viewer import Graph2dViewer

Vertex = tuple[int, int]
Edge = tuple[Vertex, Vertex]

if __name__ == '__main__':
    from algoritmia.data.iberia import iberia2d

    colors_set = list(coloring_solve(iberia2d))
    print(f"Solution with {len(colors_set)} colors.")
    colors = {}
    cols = ['red', 'green', 'blue', 'orange', 'yellow', 'brown', 'pink', 'black', 'white']
    if len(colors_set) > len(cols):
        raise Exception('Not enought colors')

    for i, s in enumerate(colors_set):
        for v in s:
            colors[v] = cols[i]

    gv = Graph2dViewer(iberia2d, canvas_width=800, canvas_height=800, vertexmode=Graph2dViewer.X_Y,
                       colors=colors, title=f"Graph coloring - {len(colors_set)} colors")
    gv.run()



from typing import *
from math import sqrt

from algoritmia.datastructures.graphs import IGraph, Vertex, Edge, WeightingFunction, Weight
from algoritmia.datastructures.queues import Fifo
from algoritmia.datastructures.prioritymaps import MinHeapMap
from algoritmia.utils import argmin, infinity

VertexTraverser = Callable[[IGraph, Vertex], Iterable[Vertex]]
EdgeTraverser = Callable[[IGraph, Vertex], Iterable[Edge]]


# Vertex traversers ----------------------------------------------------------------

def bf_vertex_traverser(graph: IGraph, v_initial: Vertex) -> Iterable[Vertex]:
    queue = Fifo()
    seen = set()
    queue.push(v_initial)
    seen.add(v_initial)
    while len(queue) > 0:
        v = queue.pop()
        yield v
        for suc in graph.succs(v):
            if suc not in seen:
                queue.push(suc)
                seen.add(suc)


def df_vertex_traverser(graph: IGraph, v_initial: Vertex) -> Iterable[Vertex]:
    def traverse_from(v: Vertex) -> Iterable[Vertex]:
        seen.add(v)
        yield v  # pre-order
        for suc in graph.succs(v):
            if suc not in seen:
                yield from traverse_from(suc)
        # yield v # post-order

    seen = set()
    yield from traverse_from(v_initial)


# Edge traversers ----------------------------------------------------------------

def bf_edge_traverser(graph: IGraph, v_initial: Vertex) -> Iterable[Edge]:
    queue = Fifo()
    seen = set()
    queue.push((v_initial, v_initial))
    seen.add(v_initial)
    while len(queue) > 0:
        u, v = queue.pop()
        yield u, v
        for suc in graph.succs(v):
            if suc not in seen:
                queue.push((v, suc))
                seen.add(suc)


def df_edge_traverser(graph: IGraph, v_initial: Vertex) -> Iterable[Edge]:
    def traverse_from(u: Vertex, v: Vertex) -> Iterable[Edge]:
        seen.add(v)
        yield u, v  # pre-order
        for suc in graph.succs(v):
            if suc not in seen:
                yield from traverse_from(v, suc)
        # yield u, v # post-order

    seen = set()
    return traverse_from(v_initial, v_initial)


# Con diccionario: O(|V|^2)
def dijkstra_edge_traverser(g: IGraph, d: WeightingFunction,
                            v_initial: Vertex) -> Iterable[Edge]:
    D: dict[Vertex, Weight] = dict((v, infinity) for v in g.V)
    D[v_initial] = 0
    bp: dict[Vertex, Vertex] = {v_initial: v_initial}
    fixed: set[Vertex] = set()
    while len(bp) > 0:  # O(|V|) veces
        v = argmin(bp.keys(), lambda u: D[u])  # O(|V|)
        fixed.add(v)
        pred_v = bp[v]
        yield pred_v, v
        del bp[v]
        for suc_v in g.succs(v):  # O(|V|)
            if suc_v not in fixed and D[v] + d(v, suc_v) < D[suc_v]:
                D[suc_v] = D[v] + d(v, suc_v)
                bp[suc_v] = v


# Con diccionario de prioridad: O(|V| + |E| log |V|)
def dijkstra_hm_edge_traverser(g: IGraph, d: WeightingFunction,
                               v_initial: Vertex) -> Iterable[Edge]:
    D: MinHeapMap[Vertex, Weight] = MinHeapMap((v, infinity) for v in g.V)  # O(|V|)
    D[v_initial] = 0
    bp: dict[Vertex, Vertex] = {v_initial: v_initial}
    fixed: set[Vertex] = set()
    while len(D) > 0:
        v, dv = D.extract_opt_item()  # O(log |V|), O(|V|) veces
        fixed.add(v)
        pred_v = bp[v]
        yield pred_v, v
        for suc_v in g.succs(v):
            if suc_v not in fixed and dv + d(v, suc_v) < D[suc_v]:
                D[suc_v] = dv + d(v, suc_v)  # O(log |V|), O(|E|) veces
                bp[suc_v] = v


# Distancia euclídea
def d_eu(u, v):
    a = u[0] - v[0]
    b = u[1] - v[1]
    return sqrt(a * a + b * b)


def dijkstra_metric_edge_traverser(g: IGraph, d: WeightingFunction,
                                   v_initial: Vertex, v_final: Vertex) -> Iterable[Edge]:
    D: dict[Vertex, Weight] = dict((v, infinity) for v in g.V)
    D[v_initial] = 0
    bp: dict[Vertex, Vertex] = {v_initial: v_initial}
    fixed: set[Vertex] = set()
    while len(bp) > 0:
        v = argmin(bp.keys(), lambda u: D[u] + d_eu(u, v_final))
        fixed.add(v)
        pred_v = bp[v]
        yield pred_v, v
        if v == v_final:
            break
        del bp[v]
        for suc_v in g.succs(v):
            if suc_v not in fixed and D[v] + d(v, suc_v) < D[suc_v]:
                D[suc_v] = D[v] + d(v, suc_v)
                bp[suc_v] = v


if __name__ == '__main__':
    from algoritmia.datastructures.graphs import UndirectedGraph

    my_edges = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)), ((0, 1), (0, 2)),
                ((2, 0), (1, 0)), ((2, 1), (2, 2)), ((2, 2), (2, 3)), ((0, 1), (1, 1)),
                ((0, 2), (1, 2)), ((0, 3), (1, 3)), ((1, 1), (2, 1)), ((1, 2), (2, 2))]

    my_graph = UndirectedGraph(E=my_edges)
    my_vertex_source = (0, 0)

    print('vertex_traverser_breadthfirst:', list(bf_vertex_traverser(my_graph, my_vertex_source)))
    print('edge_traverser_breadthfirst', list(bf_edge_traverser(my_graph, my_vertex_source)))

    print('vertex_traverser_depthfirst', list(df_vertex_traverser(my_graph, my_vertex_source)))
    print('edge_traverser_depthfirst', list(df_edge_traverser(my_graph, my_vertex_source)))

    # ---------------

    from algoritmia.data.mallorca import Mallorca, km

    print('edge_traverser_dijkstra', list(dijkstra_edge_traverser(Mallorca, km, 'Andratx')))



from typing import *

from algoritmia.datastructures.graphs import UndirectedGraph

T = TypeVar('T')


# agrupa los vertices que pueden compartir color
def coloring_solve(g: UndirectedGraph) -> list[set[T]]:
    solution = []

    # Creamos un nuevo conjunto con todos los vértices del grafo
    vertex_set = set(g.V)

    # Mientras queden vértices en el conjunto que no tengan grupo
    while len(vertex_set) > 0:
        # Creamos un nuevo grupo
        grupo = set()
        # Recorremos todos los vertices sin grupo
        for v in vertex_set:
            if not any(suc in grupo for suc in g.succs(v)):
                grupo.add(v)
        # Cerramos el grupo, lo añadimos a la solución y
        # quitamos de vertex_set los vértices que hemos metido en el nuevo grupo
        solution.append(grupo)
        vertex_set -= grupo

    return solution


if __name__ == '__main__':
    graph = UndirectedGraph(E=[(0, 1), (0, 2), (0, 3), (1, 3), (1, 4), (2, 3), (2, 5),
                               (3, 4), (3, 5), (3, 6), (4, 7), (5, 6), (5, 8), (6, 7),
                               (6, 8), (6, 9), (7, 9), (8, 9)])
    print(coloring_solve(graph))



from typing import *

from algoritmia.datastructures.graphs import UndirectedGraph, Vertex

from traversers import EdgeTraverser


def spanning_forest(g: UndirectedGraph[Vertex], traverser: EdgeTraverser) -> Iterable[UndirectedGraph]:
    pending_vertices = set(g.V)
    while len(pending_vertices) > 0:
        u = pending_vertices.pop()
        e_gen = iter(traverser(g, u))
        next(e_gen)  # skip phantom edge
        visited_edges = list(e_gen)
        pending_vertices -= set([v for (u, v) in visited_edges])
        yield UndirectedGraph(E=list(visited_edges))


if __name__ == '__main__':
    from traversers import bf_edge_traverser, df_edge_traverser

    edges = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)), ((2, 0), (2, 1)),
             ((2, 2), (2, 3)), ((0, 1), (1, 1)), ((0, 2), (1, 2)), ((0, 3), (1, 3)),
             ((1, 2), (2, 2))]

    my_graph = UndirectedGraph(E=edges)

    print("Breath first:")
    graphs_bf = list(spanning_forest(my_graph, bf_edge_traverser))
    for g in graphs_bf:
        print(g)

    print("Depth first:")
    graphs_df = list(spanning_forest(my_graph, df_edge_traverser))
    for g in graphs_df:
        print(g)



# coding: latin1

import unittest

from algoritmia.datastructures.mergefindsets import MergeFindSet


class TestMFset(unittest.TestCase):
    def setUp(self):
        self.mf1 = MergeFindSet()
        self.mf2 = MergeFindSet(((i,) for i in range(10)))

    def test_mfsets(self):
        for i in range(10):
            self.mf1.add(i)
        for i in range(10):
            self.assertEqual(self.mf1.find(i), i)
            self.assertEqual(self.mf2.find(i), i)
        for i in range(0, 10, 2):
            self.mf1.merge(i, i + 1)
            self.mf2.merge(i, i + 1)
        for i in range(0, 10, 2):
            self.assertEqual(self.mf1.find(i), self.mf1.find(i + 1))
            self.assertEqual(self.mf2.find(i), self.mf2.find(i + 1))
        for i in range(0, 10 - 3, 4):
            self.mf1.merge(i, i + 3)
            self.mf2.merge(i, i + 3)
        for i in range(0, 10 - 4, 4):
            self.assertEqual(self.mf1.find(i), self.mf1.find(i + 1))
            self.assertEqual(self.mf1.find(i), self.mf1.find(i + 2))
            self.assertEqual(self.mf1.find(i), self.mf1.find(i + 3))
            self.assertEqual(self.mf2.find(i), self.mf2.find(i + 1))
            self.assertEqual(self.mf2.find(i), self.mf2.find(i + 2))
            self.assertEqual(self.mf2.find(i), self.mf2.find(i + 3))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
Visor de grafos de dependencias y trellis de programacion dinámica
@author: David Llorens dllorens@uji.es

02/12/2021: Version 1
"""

from typing import *
from math import sin, cos, pi, sqrt, atan2
import tkinter

from easypaint import EasyPaint

Num = Union[int, float]
Label = str


class NodeStyle:
    def __init__(self, color: str = 'black', width: int = 1, fontsize: int = 10, fx: float = 1.0, fy: float = 1.0):
        self.color = color
        self.width = width
        self.fontsize = fontsize
        self.fx = fx
        self.fy = fy

    def _key(self) -> tuple[str, int]:
        return self.color, self.width

    def __hash__(self) -> int:
        return hash(self._key)

    def __eq__(self, other: 'NodeStyle') -> bool:
        if not isinstance(other, NodeStyle):
            return False
        return self._key() == other._key()


class EdgeStyle(NodeStyle):
    def __init__(self, color: str = 'black', width: int = 1, fontsize: int = 10, fx: float = 1.0, fy: float = 1.0,
                 pos: float = 0.5):
        super().__init__(color, width, fontsize, fx, fy)
        self.pos = pos

    def __key(self) -> tuple[tuple[str, int], float]:
        return super()._key(), self.pos


Pos = tuple[Num, Num]
Node = tuple[Pos, Label, NodeStyle]
Edge = tuple[Node, Node, Label, NodeStyle]


# -------------------------------------------------------------------------------------

def dist(p1, p2):
    return sqrt((p1[0] - p2[0]) * (p1[0] - p2[0]) + (p1[1] - p2[1]) * (p1[1] - p2[1]))


def find_intersec_point(cx, cy, radius, p1x, p1y, p2x, p2y):
    dx = p2x - p1x
    dy = p2y - p1y

    A = dx * dx + dy * dy
    B = 2 * (dx * (p1x - cx) + dy * (p1y - cy))
    C = (p1x - cx) * (p1x - cx) + (p1y - cy) * (p1y - cy) - radius * radius

    det = B * B - 4 * A * C
    if A <= 0.0000001 or det < 0:
        raise Exception("No real solutions")
    if det == 0:
        # One solution.
        t = -B / (2 * A)
        return (p1x + t * dx, p1y + t * dy), None
    else:
        # Two solutions.
        t1 = (-B + sqrt(det)) / (2 * A)
        t2 = (-B - sqrt(det)) / (2 * A)
        return (p1x + t1 * dx, p1y + t1 * dy), (p1x + t2 * dx, p1y + t2 * dy)


# -------------------------------------------------------------------------------------

class DPViewer(EasyPaint):
    current = 0

    def __init__(self, ledges: list[list[Edge]], lnodes: list[set[Node]], is_directed=True,
                 canvas_width: int = 400, canvas_height: int = 300, margin: int = 15,
                 axes_names=('', ''),
                 background='white',
                 node_size: Optional[float] = None,
                 titles=None,
                 swap: bool = False):
        super().__init__()
        self.node_size = node_size
        self.margin = margin
        self.ledges = ledges
        self.edges = ledges[self.current]
        self.lnodes = lnodes
        self.nodes = self.lnodes[self.current]
        self.titles = titles
        self.axes_names = axes_names
        self.swap = swap

        self.is_directed = is_directed

        self.num_x = len(set(p[0] for (p, _, _) in self.nodes))
        self.num_y = len(set(p[1] for (p, _, _) in self.nodes))

        self.min_x = min(p[0] for (p, _, _) in self.nodes)
        self.max_x = max(p[0] for (p, _, _) in self.nodes)

        self.min_y = min(p[1] for (p, _, _) in self.nodes)
        self.max_y = max(p[1] for (p, _, _) in self.nodes)

        self.title = 'Graph Viewer' if titles is None else titles[0]
        self.background = background
        self.size = canvas_width, canvas_height
        if self.swap:
            self.coordinates = (0, self.size[1], self.size[0], 0)

        w, h = self.size
        w -= self.margin
        h -= self.margin
        sizex = self.max_x - self.min_x + 1
        sizey = self.max_y - self.min_y + 1
        self.cell_size = cell_size = min(w / sizex, h / sizey)

        self.m = ((w - cell_size * sizex) / 2 - self.min_x * cell_size + self.margin / 2,
                  (h - cell_size * sizey) / 2 - self.min_y * cell_size + self.margin / 2)
        if self.node_size is None:
            self.node_size = (w * h / len(self.nodes)) ** 0.5 / 10  # cell_size / 8

    def on_key_press(self, keysym):
        if keysym in ['Return', 'Escape']:
            self.close()
        if keysym.upper() == 'SPACE':
            self.current = (self.current + 1) % len(self.ledges)
            self.edges = self.ledges[self.current]
            self.title = self.titles[self.current]
            self.erase('arrow')
            for edge in self.edges:
                self.draw_arrow(edge)
            self.nodes = self.lnodes[self.current]
            self.erase('vertex')
            for node in self.nodes:
                self.draw_vertex(node)
        if keysym.upper() == 'P':
            self.save_eps("saved.eps")

    def draw_arrow(self, edge, tag='arrow'):
        cell_size = self.cell_size
        m = self.m
        (u, _, _), (v, _, _), label, style = edge
        width = style.width

        if not self.is_directed:
            self.create_line((u[0] + 0.5) * cell_size + m[0], (u[1] + 0.5) * cell_size + m[1],
                             (v[0] + 0.5) * cell_size + m[0], (v[1] + 0.5) * cell_size + m[1])
        else:
            p1, p2 = find_intersec_point(v[0], v[1], (self.node_size + width + 2) / self.cell_size,
                                         u[0], u[1], v[0], v[1])
            _, (x2, y2) = min((dist(u, p1), p1), (dist(u, p2), p2))

            p1, p2 = find_intersec_point(u[0], u[1], (self.node_size + width + 2) / self.cell_size,
                                         v[0], v[1], u[0], u[1])
            _, (x1, y1) = min((dist(v, p1), p1), (dist(v, p2), p2))

            self._draw_arrow(x1, y1, x2, y2, label, style, tag=tag)

    def _draw_arrow(self, x1, y1, x2, y2, label, style, tag='arrow'):
        cell_size = self.cell_size
        m = self.m
        color = style.color
        width = style.width
        fs = style.fontsize
        pos = style.pos

        # x1, y1 = u
        l2 = self.node_size

        phi = 30 * pi / 180
        l1 = ((x1 - x2) ** 2 + (y1 - y2) ** 2) ** 0.5
        l2 /= (self.cell_size * 4)

        x3 = x2 + l2 / l1 * ((x1 - x2) * cos(phi) + (y1 - y2) * sin(phi))
        x4 = x2 + l2 / l1 * ((x1 - x2) * cos(phi) - (y1 - y2) * sin(phi))

        y3 = y2 + l2 / l1 * ((y1 - y2) * cos(phi) - (x1 - x2) * sin(phi))
        y4 = y2 + l2 / l1 * ((y1 - y2) * cos(phi) + (x1 - x2) * sin(phi))

        self.create_line((x1 + 0.5) * cell_size + m[0], (y1 + 0.5) * cell_size + m[1],
                         (x2 + 0.5) * cell_size + m[0], (y2 + 0.5) * cell_size + m[1],
                         color, width=width, capstyle=tkinter.ROUND, tag=tag)
        self.create_line((x2 + 0.5) * cell_size + m[0], (y2 + 0.5) * cell_size + m[1],
                         (x3 + 0.5) * cell_size + m[0], (y3 + 0.5) * cell_size + m[1],
                         color, width=width, capstyle=tkinter.ROUND, tag=tag)
        self.create_line((x2 + 0.5) * cell_size + m[0], (y2 + 0.5) * cell_size + m[1],
                         (x4 + 0.5) * cell_size + m[0], (y4 + 0.5) * cell_size + m[1],
                         color, width=width, capstyle=tkinter.ROUND, tag=tag)
        if pos >= 0:
            def x(t): return t * x1 + (1 - t) * x2

            def y(t): return t * y1 + (1 - t) * y2

            xm, ym = x(pos), y(pos)
            xr, yr = x2 - x1, y2 - y1
            a = atan2(yr, xr)

            ix = 0
            if abs(a) < 0.01 or abs(abs(a) - pi) < 0.01:
                anchor = 'S'
            elif abs(abs(a) - pi / 2) < 0.01:
                anchor = 'E'
                ix = 2
            elif 0 <= a <= pi / 2 or -pi <= a <= -pi / 2:
                anchor = 'SE' if not self.swap else 'SW'
            else:
                anchor = 'SW' if not self.swap else 'SE'

            self.create_text((xm + 0.5) * self.cell_size + self.m[0] - ix,
                             (ym + 0.5) * self.cell_size + self.m[1],
                             label, fs, color=color, anchor=anchor, tag=tag)

    def draw_vertex(self, node: Node, tag='vertex'):
        u, label, style = node
        fill = style.color
        width = style.width
        fs = style.fontsize
        self.create_filled_circle((u[0] + 0.5) * self.cell_size + self.m[0],
                                  (u[1] + 0.5) * self.cell_size + self.m[1],
                                  self.node_size, color='black', width=width, fill=fill, tag=tag)
        if label is not None:
            ls = label.split('\n')
            if len(ls) == 1:
                label = ls[0]
                self.create_text((u[0] + 0.5) * self.cell_size + self.m[0],
                                 (u[1] + 0.5) * self.cell_size + self.m[1],
                                 label, fs, color='black', tag=tag)
            if len(ls) == 2:
                label1, label2 = ls
                self.create_text((u[0] + 0.5) * self.cell_size + self.m[0],
                                 (u[1] + 0.5) * self.cell_size + self.m[1],
                                 label1, fs, anchor="S", color='black', tag=tag)
                self.create_text((u[0] + 0.5) * self.cell_size + self.m[0],
                                 (u[1] + 0.5) * self.cell_size + self.m[1] - 0.05 * self.cell_size,
                                 label2, fs, anchor="N", color='blue', tag=tag)

    def main(self):
        for edge in self.edges:
            self.draw_arrow(edge)
        for node in self.nodes:
            self.draw_vertex(node)

        arrow_style = EdgeStyle(color='black', fontsize=12, width=1, pos=0.5)
        if self.swap:
            lh, lv = self.axes_names[1], self.axes_names[0]
        else:
            lh, lv = self.axes_names[0], self.axes_names[1]
        kk = 2 * self.node_size / self.cell_size
        if lv != '':
            self._draw_arrow(-kk, 0, -kk, (self.size[1] - 2 * self.m[1]) / self.cell_size - 1, lv, arrow_style,
                             tag='axe')
        if lh != '':
            self._draw_arrow(0, -kk, (self.size[0] - 2 * self.m[0]) / self.cell_size - 1, -kk, lh, arrow_style,
                             tag='axe')


if __name__ == '__main__':
    node_style = NodeStyle(color='palegreen', width=1, fontsize=12)
    node_styleA = NodeStyle(color='pink', width=2, fontsize=12)
    arrow_style = EdgeStyle(color='black', width=1, fontsize=12, pos=0.5)
    arrow_style2 = EdgeStyle(color='blue', width=2, fontsize=12, pos=0.5)

    a = ((10, 1), 'A', node_styleA)
    b = ((15, 4), 'B', node_styleA)
    c = ((10, 7), 'C', node_style)
    d = ((20, 4), 'D', node_style)
    nodes = {a, b, c, d}

    edges = [(a, b, '7', arrow_style2), (c, a, '1', arrow_style), (b, d, '-2', arrow_style), (c, b, '4', arrow_style)]
    edges2 = [(b, a, '-7', arrow_style2), (a, c, '-1', arrow_style), (d, b, '2', arrow_style),
              (b, c, '-4', arrow_style)]

    viewer = DPViewer([edges, edges2], [nodes, nodes],
                      titles=["Grafo A (pulsa 'Space')", "Grafo B (pulsa 'Space')"],
                      canvas_width=400, canvas_height=300)
    viewer.run()



from typing import *

from algoritmia.datastructures.graphs import UndirectedGraph, Vertex

from traversers import VertexTraverser

CC = list[Vertex]  # Connected Component


def connected_components(g: UndirectedGraph[Vertex],
                         vertex_traverser: VertexTraverser) -> Iterable[CC]:
    pending_vertices = set(g.V)
    while len(pending_vertices) > 0:
        u = pending_vertices.pop()
        visited_vertices = list(vertex_traverser(g, u))
        pending_vertices -= set(visited_vertices)
        yield visited_vertices


if __name__ == '__main__':
    from traversers import bf_vertex_traverser, df_vertex_traverser

    edges = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)), ((2, 0), (2, 1)),
             ((2, 2), (2, 3)), ((0, 1), (1, 1)), ((0, 2), (1, 2)), ((0, 3), (1, 3)),
             ((1, 2), (2, 2))]

    my_graph = UndirectedGraph(E=edges)

    ccs_bf = list(connected_components(my_graph, bf_vertex_traverser))
    print("Breath first::", ccs_bf)

    ccs_df = list(connected_components(my_graph, df_vertex_traverser))
    print("Depth first:", ccs_df)



from dataclasses import dataclass
from typing import *

from algoritmia.datastructures.graphs import UndirectedGraph, WeightingFunction
from algoritmia.schemes.bt_scheme import DecisionSequence, bt_solve
from algoritmia.schemes.bt_scheme import ScoredDecisionSequence, bt_min_solve

Decision = TypeVar('Decision')
Solution = tuple[Decision, ...]


def hamiltoniancycle_solve(graph: UndirectedGraph) -> Iterable[Solution]:
    class HamiltonianCycleDS(DecisionSequence):
        def is_solution(self):
            ds = self.decisions()
            return len(ds) == len(graph.V) and ds[0] in graph.succs(ds[-1])

        def successors(self) -> Iterable['HamiltonianCycleDS']:
            ds = self.decisions()
            if len(ds) < len(graph.V):
                for v in graph.succs(ds[-1]):
                    if v not in ds:
                        yield self.add_decision(v)

    v_initial = next(iter(graph.V))  # Para coger un elemento de un conjunto sin sacarlo
    initial_ds = HamiltonianCycleDS().add_decision(v_initial)
    return bt_solve(initial_ds)


Solution = tuple[float, tuple[Decision, ...]]
State = tuple[int, tuple[int, ...]]
Score = float


def hamiltoniancycle_opt_solve(graph: UndirectedGraph, wf: WeightingFunction):
    @dataclass
    class Extra:
        ls: int = 0

    class HamiltonianCycleDS(ScoredDecisionSequence):
        def is_solution(self):
            ds = self.decisions()
            return len(ds) == len(graph.V) and (ds[-1], ds[0]) in graph.E

        def solution(self) -> Solution:
            return self.score(), self.decisions()

        def successors(self) -> Iterable['HamiltonianCycleDS']:
            ds = self.decisions()
            if len(ds) < len(graph.V):
                for v in graph.succs(ds[-1]):
                    if v not in ds:
                        new_extra = Extra(self.extra.ls + wf(ds[-1], v))
                        yield self.add_decision(v, new_extra)

        def state(self) -> State:
            ds = list(self.decisions())
            return ds[-1], tuple(sorted(ds))

        def score(self) -> Score:
            if len(self) < len(graph.V): return self.extra.ls
            ds = self.decisions()
            if ds[0] in graph.succs(ds[-1]):
                return self.extra.ls + wf(ds[0], ds[-1])
            return float('infinity')

    v_initial = next(iter(graph.V))  # Para coger un elemento de un conjunto sin sacarlo
    initial_ds = HamiltonianCycleDS(Extra()).add_decision(v_initial, Extra())
    return bt_min_solve(initial_ds)


# Programa principal ---------------------------
if __name__ == "__main__":
    G = UndirectedGraph(E=[(0, 2), (0, 3), (0, 9), (1, 3), (1, 4), (1, 8), (2, 3), (2, 5), (3, 4),
                           (3, 6), (4, 7), (5, 6), (5, 8), (6, 7), (6, 8), (6, 9)])
    d: dict[tuple[int, int], int] = {}
    for (u, v) in G.E:
        d[u, v] = abs(u - v)
    wf = WeightingFunction((e for e in d.items()), symmetrical=True)

    print("hamiltoniancycle_solve:")
    for solution in hamiltoniancycle_solve(G):
        print(solution)
    print()
    print("hamiltoniancycle_opt_solve:")
    for solution in hamiltoniancycle_opt_solve(G, wf):
        print(solution)



from dataclasses import dataclass
from random import seed, randint
from typing import *

from algoritmia.schemes.bab_scheme import BoundedDecisionSequence, bab_max_solve

Decision = int  # 0 o 1
Solution = tuple[int, int, tuple[Decision,...]]  # (value, weight, decisions)


def knapsack_bab_solve(weights: list[int], values: list[int], capacity: int) -> Solution:
    @dataclass
    class Extra:
        weight: int = 0
        value: int = 0

    class KnapsackBabDS(BoundedDecisionSequence):
        # Coge TODOS los objetos pendientes
        def calculate_opt_bound(self) -> int:
            return self.extra.value + sum(values[len(self):])

        # No coge NINGUNO de los objetos pendientes
        def calculate_pes_bound(self) -> int:
            return self.extra.value

        def is_solution(self) -> bool:
            return len(self) == len(values)

        def solution(self) -> Solution:
            return self.extra.value, self.extra.weight, self.decisions()

        def successors(self) -> Iterable["KnapsackBabDS"]:
            n = len(self)
            if n < len(values):
                if weights[n] <= capacity - self.extra.weight:
                    new_extra = Extra(self.extra.weight + weights[n], self.extra.value + values[n])
                    yield self.add_decision(1, new_extra)
                yield self.add_decision(0, self.extra)

    initial_ds = KnapsackBabDS(Extra())
    return bab_max_solve(initial_ds)


def knapsack_bab_solve2(weights: list[int], values: list[int], capacity: int):
    @dataclass
    class Extra:
        weight: int = 0
        value: int = 0

    class KnapsackBabDS(BoundedDecisionSequence):
        # IMPLEMENTAR: resolver mochila continua para los objetos que quedan (tema 3)
        def calculate_opt_bound(self) -> int:
            value = self.extra.value
            weight = self.extra.weight
            for i in range(len(self), len(weights)):
                f = min(1, (capacity - weight) / weights[i])
                weight += f * weights[i]
                value += f * values[i]
                if f < 1:
                    break
            return value

        # IMPLEMENTAR: modificación del optimista sin fraccionar
        def calculate_pes_bound(self) -> int:
            value = self.extra.value
            weight = self.extra.weight
            for i in range(len(self), len(weights)):
                if (capacity - weight) >= weights[i]:
                    weight += weights[i]
                    value += values[i]
            return value

        def is_solution(self) -> bool:
            return len(self) == len(values)

        def solution(self) -> Solution:
            return self.extra.value, self.extra.weight, self.decisions()

        def successors(self) -> Iterable["KnapsackBabDS"]:
            n = len(self)
            if n < len(values):
                if weights[n] <= capacity - self.extra.weight:
                    new_extra = Extra(self.extra.weight + weights[n], self.extra.value + values[n])
                    yield self.add_decision(1, new_extra)
                yield self.add_decision(0, self.extra)

    initial_ps = KnapsackBabDS(Extra())
    return bab_max_solve(initial_ps)


def sorted_by_dec_ratio(w_old, v_old):
    idxs = sorted(range(len(w_old)), key=lambda i: -v_old[i] / w_old[i])
    w_new = [w_old[i] for i in idxs]
    v_new = [v_old[i] for i in idxs]
    return w_new, v_new


def create_knapsack_problem(num_objects):
    seed(5)
    w_new = [randint(10, 100) for _ in range(num_objects)]
    v_new = [w_new[i] * randint(1, 4) for i in range(num_objects)]
    capacity = int(sum(w_new) * 0.3)
    weights, values = sorted_by_dec_ratio(w_new, v_new)
    return weights, values, capacity


# Main program -------------------------------------------------------
if __name__ == "__main__":
    # W, V, C = [1, 5, 6, 2, 6], [1, 2, 3, 4, 2], 10  # Solution: value = 8,    weight = 9,   decisions = (1, 0, 1, 1, 0))
    # W, V, C = create_knapsack_problem(20)           # Solution: value = 1118, weight = 344, decisions = (1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0)
    W, V, C = create_knapsack_problem(
        35)  # Solution: value = 1830, weight = 543, decisions = (1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

    print("Solution: ", knapsack_bab_solve2(W, V, C))  # Solution:  (8, 9, (1, 0, 1, 1, 0))



from enum import Enum

LParams = tuple[int, int]
Score = int
Decision = str
Vertex = tuple[int, int]


class RecMode(Enum):
    Classic = 0
    Optimized = 1


def edit_distance(mode: RecMode, s: str, t: str) -> tuple[Score, list[Decision]]:
    def D_classic(m: int, n: int) -> Score:
        if m == 0 and n == 0:
            return 0
        if (m, n) not in mem:
            if n == 0:
                mem[m, n] = D_classic(m - 1, n) + 1, (m - 1, n), 'D'
            elif m == 0:
                mem[m, n] = D_classic(m, n - 1) + 1, (m, n - 1), 'I'
            else:
                c, d = (0, '-') if s[m - 1] == t[n - 1] else (1, 'S')
                mem[m, n] = min((D_classic(m - 1, n) + 1, (m - 1, n), 'D'),
                                (D_classic(m, n - 1) + 1, (m, n - 1), 'I'),
                                (D_classic(m - 1, n - 1) + c, (m - 1, n - 1), d))
        return mem[m, n][0]

    def D_optimized(m: int, n: int) -> Score:
        if m == 0 and n == 0:
            return 0
        if (m, n) not in mem:
            if n == 0:
                mem[m, n] = D_classic(m - 1, n) + 1, (m - 1, n), 'D'
            elif m == 0:
                mem[m, n] = D_classic(m, n - 1) + 1, (m, n - 1), 'I'
            else:
                if s[m - 1] == t[n - 1]:
                    mem[m, n] = D_classic(m - 1, n - 1), (m - 1, n - 1), '-'
                else:
                    mem[m, n] = min((D_classic(m - 1, n) + 1, (m - 1, n), 'D'),
                                    (D_classic(m, n - 1) + 1, (m, n - 1), 'I'),
                                    (D_classic(m - 1, n - 1) + 1, (m - 1, n - 1), 'S'))
        return mem[m, n][0]

    mem: dict[LParams, tuple[Score, LParams, Decision]] = {}
    if mode == RecMode.Classic:
        score = D_classic(len(s), len(t))
    else:
        score = D_optimized(len(s), len(t))

    sol = []
    m, n = len(s), len(t)
    while (m, n) != (0, 0):
        _, (m, n), dec = mem[m, n]
        sol.append(dec)
    sol.reverse()

    return score, sol


# ------------------------------------------------------------------------------------------------------

if __name__ == '__main__':
    mode = RecMode.Classic
    my_source, my_target = 'costa', 'casa'
    print(f'Edit distance fom "{my_source}" to "{my_target}":')
    print(f'\tMode: {mode}')
    res_score, res_decisions = edit_distance(mode, my_source, my_target)
    print(f'\tScore: {res_score}\n\tDecisions: {res_decisions}')
